
package simula.random;

import java.util.Random;

/**
 * This class implements a random generator with a Bernoulli distribution.
 * It uses the standard random engine of java <code>java.util.Random</code> as basis.
 * @see java.util.Random
 * @see Generator
 */
public class Bernoulli extends Generator {

	/**
	 * This is the binomial coefficient: the success probability.
	 */
	private double p;
  

	/**
	 * Creates a new Bernoulli-distributed generator using the current time as
	 * seed and the given success probability to model the distribution.
	 * @param seed Seed for the pseudo-random generator.
	 * @param p    Parameter for the Bernoulli generator (success probability).
	 */
	public Bernoulli(double p) {
		random = new RandomEngineJava();
		this.p = p;
	}
	/**
	 * Creates a new Bernoulli-distributed generator using the given
	 * seed to initialize it and the given success probability to
	 * model it.
	 * @param seed Seed for the pseudo-random generator.
	 * @param p    Parameter for the Bernoulli generator (success probability).
	 */  
	public Bernoulli(long seed, double p) {
		random = new RandomEngineJava(seed);
		this.p = p;
	}
	/**
	 * Creates a new Bernoulli-distributed generator using the given pseudo-random
	 * engine and the given success probability to model the distribution.
	 * @param e Pseudo-random engine to use.
	 * @param p Parameter for the Bernoulli generator (success probability).
	 */
	public Bernoulli(RandomEngine e, double p) {
		random = e;
		this.p = p;
	}
	/**
	 * Returns the value of the density in the given point.
	 * @param x Point to calculate the density of.
	 * @return the value of the density in the given point.
	 */
	public double density(double x) {
		double res=0;

		if (x == 1)
			res = p;
		else
			res = 1-p;
		return res; 
	}
	/**
	 * Draws a boolean value according to the Bernoulli distribution.
	 * @return An instance of Value that encapsulates a boolean value.
	 * @see BooleanValue
	 * @see Value
	 */ 
	public Value draw() {
		return new BooleanValue( random.nextDouble() <= p );
	} 
	/**
	 * Sets a different parameter for this generator.
	 * @param p New success probability.
	 */  
	public void setParam(double p) {
		this.p = p;
	}
	/**
	 * Returns the type of the extracted value.
	 */
	public int type() {
		return BOOLEAN_TYPE;
	}
}